/******************************************************************************

Coral Tree dna/connect.cc
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include <gpl/gpl.h>
#include <wba/wba.h>
#include <dna/dna.h>

/******************************************************************************
class DNA_Connection
******************************************************************************/

/***********************************************************************
void DNA_Connection::Detach(void)
***********************************************************************/
void DNA_Connection::Detach(void)
{
	// If this is a valid reference and it is the last one to this
	// implementation (presuming that the user has not sent the same
	// implementation to more than one binding with SetImp())...
	if (GPL_Reference::IsValid() && !GPL_Reference::IsShared())
	{
		// Disconnect the connection.
		if (DNA_CHECK_STATE(DNA_CONNECTION_CONNECTED))
		{
			Disconnect();
		}
	}
	DNA_Binding::Detach();
}

/***********************************************************************
GPL_Boolean DNA_Connection::IsConnected(void)
***********************************************************************/
GPL_Boolean DNA_Connection::IsConnected(void)
{
	// Verify the existence of a connection.
	register DNA_BindingRep *r = (DNA_BindingRep *) rep;
	if ((DNA_BindingRep *) GPL_BAD_ADDRESS == r)
	{
		return FALSE;
	}
	if (!DNA_CHECK_STATE(DNA_CONNECTION_VALID | DNA_CONNECTION_OPEN |
							DNA_CONNECTION_BOUND | DNA_CONNECTION_CONNECTED))
	{
		error.Set(DNA_ERROR_BAD_STATE);
		return FALSE;
	}
	if (DNA_NO_IMP)
	{
		error.Set(DNA_ERROR_BAD_IMPLEMENTATION);
		return FALSE;
	}

	// Perform a do-nothing read to allow the socket to see
	// if its connection has been reset by the other side.
	char buffer[1];
	unsigned long length = 0;
	if (r->imp->Read(buffer, 0, 0, length, error))
	{
		return TRUE;
	}
	else if (DNA_ERROR_NO_DATA_AVAILABLE == error.GetType())
	{
		error.Set(DNA_ERROR_NONE);
		return TRUE;
	}
	return FALSE;
}

/***********************************************************************
GPL_Boolean DNA_Connection::Accept(DNA_Connection & c)
***********************************************************************/
GPL_Boolean DNA_Connection::Accept(DNA_Connection & c)
{
	GPL_ENSURE(DNA_BindingRep, r);
	if (!c.GPL_Reference::IsValid())
	{
		c.rep = new DNA_BindingRep();
		if ((DNA_BindingRep *) GPL_BAD_ADDRESS == c.rep)
		{
			gplLog("DNA ERROR", "Out of memory");
			return FALSE;
		}
	}
	register DNA_BindingRep *s = (DNA_BindingRep *) c.rep;
	if (!DNA_CHECK_STATE(DNA_CONNECTION_VALID | DNA_CONNECTION_OPEN | DNA_CONNECTION_BOUND))
	{
		error.Set(DNA_ERROR_BAD_STATE);
		DNA_LOG_ERROR;
		return FALSE;
	}

	if (DNA_NO_IMP)
	{
		error.Set(DNA_ERROR_BAD_IMPLEMENTATION);
		DNA_LOG_ERROR;
		return FALSE;
	}

	register DNA_ConnectionImp *newImp =
		(DNA_ConnectionImp *) r->imp->NewInstance();
	if (!((DNA_ConnectionImp *) r->imp)->Accept(newImp, error))
	{
		switch(error.GetType())
		{
		case DNA_ERROR_INTERRUPTED_CALL:
			break;
		case DNA_ERROR_PROTOCOL_ERROR:
			break;
		case DNA_ERROR_BLOCK_IMPENDING:
			break;
		case DNA_ERROR_BAD_FAMILY:
			// Falls through...
		case DNA_ERROR_CONNECTION_IMPENDING:
			// Falls through...
		case DNA_ERROR_CONNECTION_REFUSED:
			// Falls through...
		case DNA_ERROR_CONNECTION_PENDING:
			// Falls through...
		case DNA_ERROR_HANDLE_CONNECTED:
			// Falls through...
		case DNA_ERROR_NETWORK_UNREACHABLE:
			// Falls through...
		case DNA_ERROR_TIME_OUT:
			// Falls through...
		case DNA_ERROR_CONNECTION_RESET:
			// Falls through...
		default:
			break;
		}
		s->state = 0;
		DNA_LOG_ERROR;
		return FALSE;
	}

	s->state = DNA_CONNECTION_VALID | DNA_CONNECTION_OPEN |
				DNA_CONNECTION_BOUND | DNA_CONNECTION_CONNECTED;
	c.SetImp(newImp);
	return TRUE;
}

/***********************************************************************
GPL_Boolean DNA_Connection::Connect(void)
***********************************************************************/
GPL_Boolean DNA_Connection::Connect(void)
{
	GPL_ENSURE(DNA_BindingRep, r);
	if (!DNA_CHECK_STATE(DNA_CONNECTION_VALID | DNA_CONNECTION_OPEN |
						DNA_CONNECTION_BOUND))
	{
		error.Set(DNA_ERROR_BAD_STATE);
		DNA_LOG_ERROR;
		return FALSE;
	}

	if (DNA_NO_IMP)
	{
		error.Set(DNA_ERROR_BAD_IMPLEMENTATION);
		DNA_LOG_ERROR;
		return FALSE;
	}

	if (!((DNA_ConnectionImp *) r->imp)->Connect(error))
	{
		switch(error.GetType())
		{
		case DNA_ERROR_NO_PERMISSION:
			// Falls through...
		case DNA_ERROR_ADDRESS_IN_USE:
			// Falls through...
		case DNA_ERROR_ADDRESS_UNAVAILABLE:
			// Falls through...
		case DNA_ERROR_BAD_PROTOCOL:
			// Falls through...
		case DNA_ERROR_BAD_HANDLE:
			// Falls through...
		case DNA_ERROR_CONNECTION_REFUSED:
			// Falls through...
		case DNA_ERROR_CONNECTION_PENDING:
			// Falls through...
		case DNA_ERROR_INTERRUPTED_CALL:
			// Falls through...
		case DNA_ERROR_BAD_ADDRESS:
			// Falls through...
		case DNA_ERROR_IO_ERROR:
			DNA_CLEAR_STATE(DNA_CONNECTION_CONNECTED);
			DNA_LOG_ERROR;
			break;
		case DNA_ERROR_HANDLE_CONNECTED:
			break;
		case DNA_ERROR_NETWORK_UNREACHABLE:
			// Falls through...
		case DNA_ERROR_INSUFFICIENT_RESOURCE:
			// Falls through...
		case DNA_ERROR_TIME_OUT:
			// Falls through...
		default:
			DNA_CLEAR_STATE(DNA_CONNECTION_CONNECTED);
			DNA_LOG_ERROR;
			break;
		}
		return FALSE;
	}

	DNA_SET_STATE(DNA_CONNECTION_CONNECTED);
	return TRUE;
}

/***********************************************************************
GPL_Boolean DNA_Connection::Disconnect(void)
***********************************************************************/
GPL_Boolean DNA_Connection::Disconnect(void)
{
	GPL_ENSURE(DNA_BindingRep, r);
	if (DNA_NO_IMP)
	{
		error.Set(DNA_ERROR_BAD_IMPLEMENTATION);
		DNA_LOG_ERROR;
		return FALSE;
	}

	if (((DNA_ConnectionImp *) r->imp)->Disconnect(error))
	{
		DNA_CLEAR_STATE(DNA_CONNECTION_CONNECTED);
		return TRUE;
	}

	DNA_LOG_ERROR;
	return FALSE;
}

/***********************************************************************
GPL_Boolean DNA_Connection::Read(void *buffer, const unsigned long size,
							unsigned long & length)
***********************************************************************/
GPL_Boolean DNA_Connection::Read(void *buffer, const unsigned long size,
							unsigned long & length)
{
	GPL_ENSURE(DNA_BindingRep, r);r=r;
	if (!DNA_CHECK_STATE(DNA_CONNECTION_OPEN | DNA_CONNECTION_BOUND |
						DNA_CONNECTION_CONNECTED))
	{
		error.Set(DNA_ERROR_BAD_STATE);
		DNA_LOG_ERROR;
		return FALSE;
	}

	register const GPL_Boolean success =
		DNA_Binding::Read(buffer, size, length);
	if (!success)
	{
		if (!DNA_CHECK_STATE(DNA_CONNECTION_OPEN | DNA_CONNECTION_BOUND))
		{
			DNA_CLEAR_STATE(DNA_CONNECTION_CONNECTED);
		}
		switch(error.GetType())
		{
		case DNA_ERROR_CONNECTION_RESET:
			DNA_CLEAR_STATE(DNA_CONNECTION_CONNECTED);
			DNA_LOG_ERROR;
			break;
		default:
			// NOP
			break;
		}
	}
	return success;
}

/***********************************************************************
GPL_Boolean DNA_Connection::Read(void *buffer, const unsigned long size,
							const long timeout, unsigned long & length)
***********************************************************************/
GPL_Boolean DNA_Connection::Read(void *buffer, const unsigned long size,
							const long timeout, unsigned long & length)
{
	GPL_ENSURE(DNA_BindingRep, r);r=r;
	if (!DNA_CHECK_STATE(DNA_CONNECTION_OPEN | DNA_CONNECTION_BOUND |
						DNA_CONNECTION_CONNECTED))
	{
		error.Set(DNA_ERROR_BAD_STATE);
		DNA_LOG_ERROR;
		return FALSE;
	}

	register const GPL_Boolean success =
		DNA_Binding::Read(buffer, size, timeout, length);
	if (!success)
	{
		if (!DNA_CHECK_STATE(DNA_CONNECTION_OPEN | DNA_CONNECTION_BOUND))
		{
			DNA_CLEAR_STATE(DNA_CONNECTION_CONNECTED);
		}
		switch(error.GetType())
		{
		case DNA_ERROR_CONNECTION_RESET:
			DNA_CLEAR_STATE(DNA_CONNECTION_CONNECTED);
			DNA_LOG_ERROR;
			break;
		default:
			// NOP
			break;
		}
	}
	return success;
}

/***********************************************************************
GPL_Boolean DNA_Connection::Write(void *buffer, const unsigned long size,
					unsigned long & length)
***********************************************************************/
GPL_Boolean DNA_Connection::Write(void *buffer, const unsigned long size,
					unsigned long & length)
{
	GPL_ENSURE(DNA_BindingRep, r);r=r;
	if (!DNA_CHECK_STATE(DNA_CONNECTION_OPEN | DNA_CONNECTION_BOUND |
						DNA_CONNECTION_CONNECTED))
	{
		error.Set(DNA_ERROR_BAD_STATE);
		DNA_LOG_ERROR;
		return FALSE;
	}

	register const GPL_Boolean success =
		DNA_Binding::Write(buffer, size, length);
	if (!success)
	{
		if (!DNA_CHECK_STATE(DNA_CONNECTION_OPEN | DNA_CONNECTION_BOUND))
		{
			DNA_CLEAR_STATE(DNA_CONNECTION_CONNECTED);
		}
		switch(error.GetType())
		{
		case DNA_ERROR_CONNECTION_RESET:
			DNA_CLEAR_STATE(DNA_CONNECTION_CONNECTED);
			DNA_LOG_ERROR;
			break;
		default:
			// NOP
			break;
		}
	}
	return success;
}

/***********************************************************************
GPL_Boolean DNA_Connection::Print(GPL_String & string,
							const GPL_Verbosity verbosity) const
***********************************************************************/
GPL_Boolean DNA_Connection::Print(GPL_String & string,
							const GPL_Verbosity verbosity) const
{
	if (GPL_Reference::IsValid())
	{
		register DNA_BindingRep *r = (DNA_BindingRep *) rep;
		GPL_String s;
		if ((DNA_ConnectionImp *) GPL_BAD_ADDRESS == r->imp)
		{
			string.Append("null");
		}
		else
		{
			s.Printf("%s(%p)", r->imp->GetClassName(), (void *) r->imp);
			string.Append(s);
		}
		string.Append(',');
		s.Clear();
		if (r->state & DNA_CONNECTION_VALID)
		{
			s.Join("|", "VALID");
		}
		if (r->state & DNA_CONNECTION_OPEN)
		{
			s.Join("|", "OPEN");
		}
		if (r->state & DNA_CONNECTION_BOUND)
		{
			s.Join("|", "BOUND");
		}
		if (r->state & DNA_CONNECTION_CONNECTED)
		{
			s.Join("|", "CONNECTED");
		}
		string.Append(s);
	}
	return TRUE;
}

GPL_OBJECT_IMPLEMENTATION(DNA_Connection, 0x40070003, "DNA_Connection");
